function createDomainInputField(options) {
	var domain = options.domain;
	var classname = domain.classname;
	
	options.target.html('');
	
	if (options.label !== undefined) {
		var labelDiv = $('<label></label>');
		labelDiv.attr('for', options.idPrefix+'value');
		labelDiv.html(options.label);
		
		options.target.append(labelDiv);
	}
	
	var valueContainer = $('<div></div>');

	if (classname.match(/.*\.RealDomain$/)) {
		createRealDomainInputField({
			'idPrefix' : options.idPrefix,
			'target' : valueContainer,
			'domain' : domain,
			'fillWithDefaults' : options.fillWithDefaults
		});
	} else if (classname.match(/.*\.IntegerDomain$/)) {
		createIntegerDomainInputField({
			'idPrefix' : options.idPrefix,
			'target' : valueContainer,
			'domain' : domain,
			'fillWithDefaults' : options.fillWithDefaults
		});
	} else if (classname.match(/.*\.CategoricalDomain$/)) {
		createCategoricalDomainInputField({
			'idPrefix' : options.idPrefix,
			'target' : valueContainer,
			'domain' : domain,
			'fillWithDefaults' : options.fillWithDefaults
		});
	} else if (classname.match(/.*\.BooleanDomain$/)) {
		createBooleanDomainInputField({
			'idPrefix' : options.idPrefix,
			'target' : valueContainer,
			'domain' : domain,
			'fillWithDefaults' : options.fillWithDefaults
		});
	} else {
		createUnknownDomainInputField({
			'idPrefix' : options.idPrefix,
			'target' : valueContainer,
			'domain' : domain,
			'fillWithDefaults' : options.fillWithDefaults
		})
	}
	
	options.target.append(valueContainer);
}

function createRealDomainInputField(options) {
	var field = $('<input></input>');
	field.attr('id', options.idPrefix+'value');
	field.attr('name', options.idPrefix+'value');
	field.attr('type', 'text');
	field.attr('size', 35);
	
	if (options.fillWithDefaults) {
		var defaultValue = options.domain['default']
		field.val(defaultValue);
	}
	
	options.target.html(field);
}

function createIntegerDomainInputField(options) {
	var field = $('<input></input>');
	field.attr('id', options.idPrefix+'value');
	field.attr('name', options.idPrefix+'value');
	field.attr('type', 'text');
	field.attr('size', 35);
	
	if (options.fillWithDefaults) {
		var defaultValue = options.domain['default']
		field.val(defaultValue);
	}
	
	options.target.html(field);
}

function createCategoricalDomainInputField(options) {
	var field = $('<select></select>');
	field.attr('id', options.idPrefix+'value');
	field.attr('name', options.idPrefix+'value');

	var defaultValue = options.domain['default']

	for (var i in options.domain.items) {
		var opt = $('<option></option>');
		opt.attr('value', options.domain.items[i]);
		opt.html(options.domain.items[i]);
		
		if (options.fillWithDefaults && options.domain.items[i] === defaultValue) {
			opt.attr('selected', true);
		}
		
		field.append(opt);
	}
	
	options.target.html(field);
}

function createBooleanDomainInputField(options) {
	var field = $('<select></select>');
	field.attr('id', options.idPrefix+'value');
	field.attr('name', options.idPrefix+'value');

	var defaultValue = options.domain['default']

	var trueOpt = $('<option></option>');
	trueOpt.attr('value', 'true');
	trueOpt.html('true');
		
	if (options.fillWithDefaults && defaultValue === 'true') {
		trueOpt.attr('selected', true);
	}
		
	field.append(trueOpt);
	
	var falseOpt = $('<option></option>');
	falseOpt.attr('value', 'false');
	falseOpt.html('false');
		
	if (options.fillWithDefaults && defaultValue === 'false') {
		falseOpt.attr('selected', true);
	}
		
	field.append(falseOpt);
	
	options.target.html(field);
}

function createUnknownDomainInputField(options) {
	var field = $('<input></input>');
	field.attr('id', options.idPrefix+'value');
	field.attr('name', options.idPrefix+'value');
	field.attr('type', 'text');
	field.attr('size', 75);
	
	if (options.fillWithDefaults) {
		var defaultValue = options.domain['default']
		field.val(defaultValue);
	}
	
	options.target.html(field);
}